# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import os
import shutil
try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

from cmf.monitor import MonitorDaemon, constants, schema
from cmf.monitor.statestore import StateStoreMonitor, json_parser
from cmf.monitor.tests import make_config_file

from nose.tools import assert_true, assert_equals, assert_false

class MockStateStoreMonitor(StateStoreMonitor):
  def __init__(self, daemon, registry, string):
    StateStoreMonitor.__init__(self, daemon, registry)
    self.set_string(string)
    self.set_throw_error(None)

  def set_string(self, string):
    self._string = string

  def set_throw_error(self, error):
    self._error = error

  def urlopen(self, url):
    if self._error:
      raise self._error
    return StringIO(self._string)

def test_state_store_lifecycle():
  daemon = MonitorDaemon("host", "host", agent=None)
  reg = daemon.get_registry_for_testing()
  monitor = MockStateStoreMonitor(daemon, reg, json_parser.SAMPLE)
  monitor.refresh_pid(os.getpid(), True)

  properties = """[statestore]
role_name = %s
service_name = my_impala
service_version = 4
host = localhost
webserver_port = 1234
log_dir = /path/to/logs
collect_interval = 1.0
statestore_webserver_port = 9091
"""

  # Make two different configs
  conf_dir = make_config_file(properties % ("statestore-test",))
  conf_dir2 = make_config_file(properties % ("statestore",))

  try:
    assert_equals(0, len(reg.get_registry()))

    # Start it
    monitor.refresh(conf_dir)
    assert_equals(monitor, reg.get_registry().pop())
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    assert_equals('statestore-test', role_update['rolename'])

    # Refresh with a different config
    monitor.refresh(conf_dir2)
    assert_equals(monitor, reg.get_registry().pop())
    assert_equals(1, len(reg.get_registry()))

    monitor.collect()

    # Test metrics and role update
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('statestore', role_update['rolename'])
    assert_equals('my_impala', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_true(metric_dict[schema.TOTAL_CPU_SYSTEM] >= 0.0)
    assert_true(metric_dict[schema.MEM_RSS] >= 0.0)
    assert_true(metric_dict[schema.MEM_VIRTUAL] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_GOOD,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 1, "We configured 1 dirs: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])

    # Test when the process is not running
    monitor.refresh_pid(None)
    monitor.collect()

    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('statestore', role_update['rolename'])
    assert_equals('my_impala', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 1, "We configured 1 dirs: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])

    # Test with a collection error
    monitor.refresh_pid(os.getpid(), True)
    monitor.set_throw_error(IOError("BOOM!"))
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_true(metric_dict[schema.MEM_RSS] >= 0.0)
    assert_true(metric_dict[schema.MEM_VIRTUAL] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_COMMUNICATION_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Test with some invalid JSON
    monitor.set_throw_error(None)
    monitor.set_string("THIS IS BOGUS JSON!")
    monitor.collect()
    monitor.clear_last_reported_timestamp()
    role_update = monitor.get_role_update().to_avro_dict()
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_true(metric_dict[schema.MEM_RSS] >= 0.0)
    assert_true(metric_dict[schema.MEM_VIRTUAL] >= 0.0)
    assert_equals(constants.WEB_METRICS_COLLECTION_PARSE_FAILURE,
                  metric_dict[schema.WEB_METRICS_COLLECTION_STATUS])
    assert_true(schema.WEB_METRICS_COLLECTION_DURATION in metric_dict)

    # Stop it
    monitor.stop()
    assert_equals(0, len(reg.get_registry()))
  finally:
    shutil.rmtree(conf_dir)
    shutil.rmtree(conf_dir2)

