# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import os
import shutil

from cmf.monitor import MonitorDaemon, schema
from cmf.monitor.zookeeper import ZooKeeperServerMonitor
from cmf.monitor.tests import make_config_file

from nose.tools import assert_true, assert_equals, assert_false

def test_zookeeper_server_lifecycle():
  daemon = MonitorDaemon("host", "host", agent=None)
  reg = daemon.get_registry_for_testing()
  monitor = ZooKeeperServerMonitor(daemon, reg)
  monitor.refresh_pid(os.getpid(), True)

  properties = """[server]
role_name = %s
service_name = my_zookeeper1
service_version = 4
host = localhost
log_dir = /path/to/logs
data_dir = /path/to/data
data_log_dir = /path/to/data/logs
collect_interval = 1.0
"""

  # Make two different configs
  conf_dir = make_config_file(properties % ("zookeeper-server-test",))
  conf_dir2 = make_config_file(properties % ("zookeeper-server",))

  try:
    assert_equals(0, len(reg.get_registry()))

    # Start it
    monitor.refresh(conf_dir)
    assert_equals(monitor, reg.get_registry().pop())
    monitor.collect()
    role_update = monitor.get_role_update().to_avro_dict()
    assert_equals('zookeeper-server-test', role_update['rolename'])

    # Refresh with a different config
    monitor.refresh(conf_dir2)
    assert_equals(monitor, reg.get_registry().pop())
    assert_equals(1, len(reg.get_registry()))

    monitor.collect()

    # Test metrics and role update
    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('zookeeper-server', role_update['rolename'])
    assert_equals('my_zookeeper1', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']
    assert_true(metric_dict[schema.TOTAL_CPU_USER] >= 0.0)
    assert_true(metric_dict[schema.TOTAL_CPU_SYSTEM] >= 0.0)

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 3, "We configured 3 dirs: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])
    assert_equals('/path/to/data', dir_metrics[1]['path'])
    assert_equals('/path/to/data/logs', dir_metrics[2]['path'])

    # Test when the process is not running
    monitor.refresh_pid(None)
    monitor.collect()

    role_update = monitor.get_role_update().to_avro_dict()

    assert_equals('zookeeper-server', role_update['rolename'])
    assert_equals('my_zookeeper1', role_update['service_name'])
    metric_dict = { }
    for metric_obj in role_update['metrics']:
      metric_dict[metric_obj['id']] = metric_obj['value']

    dir_metrics = role_update['dirUpdates']
    assert_true(len(dir_metrics) == 3, "We configured 3 dirs: %s" % dir_metrics)
    assert_true(os.path.ismount(dir_metrics[0]['mountpoint']))
    assert_equals('/path/to/logs', dir_metrics[0]['path'])
    assert_equals('/path/to/data', dir_metrics[1]['path'])
    assert_equals('/path/to/data/logs', dir_metrics[2]['path'])

    # Stop it
    monitor.stop()
    assert_equals(0, len(reg.get_registry()))
  finally:
    shutil.rmtree(conf_dir)
    shutil.rmtree(conf_dir2)
